@@ -3,8 +3,8 @@
 /*
     This file is part of Magnum.
 
-    Copyright © 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019
-              Vladimír Vondruš <mosra@centrum.cz>
+    Copyright © 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019,
+                2020 Vladimír Vondruš <mosra@centrum.cz>
     Copyright © 2018 Tobias Stein <stein.tobi@t-online.de>
 
     Permission is hereby granted, free of charge, to any person obtaining a
@@ -31,12 +31,14 @@
  */
 
 #include <Magnum/Trade/AbstractImporter.h>
+#include <Magnum/Trade/PhongMaterialData.h>
 
 #include "MagnumPlugins/TinyGltfImporter/configure.h"
 
 #ifndef DOXYGEN_GENERATING_OUTPUT
 namespace tinygltf {
     class Model;
+    class Value;
 }
 #endif
 
@@ -61,21 +63,12 @@ namespace Magnum { namespace Trade {
 /**
 @brief TinyGltf importer plugin
 
-@m_keywords{GltfImporter GlbImporter}
+@m_keywords{GltfImporter}
 
 Imports glTF and binary glTF using the [TinyGLTF](https://github.com/syoyo/tinygltf)
 library.
 
-This plugin depends on the @ref Trade library and is built if
-`WITH_TINYGLTFIMPORTER` is enabled when building Magnum Plugins. To use as a
-dynamic plugin, you need to load the @cpp "TinyGltfImporter" @ce plugin from
-`MAGNUM_PLUGINS_IMPORTER_DIR`. To use as a static plugin or as a dependency of
-another plugin with CMake, you need to request the `TinyGltfImporter` component
-of the `MagnumPlugins` package and link to the
-`MagnumPlugins::TinyGltfImporter` target. See @ref building-plugins,
-@ref cmake-plugins and @ref plugins for more information.
-
-This plugin provides `GltfImporter` and `GlbImporter` plugins.
+This plugin provides the `GltfImporter` plugin.
 
 @m_class{m-block m-success}
 
@@ -88,20 +81,59 @@ This plugin provides `GltfImporter` and `GlbImporter` plugins.
     @m_class{m-label m-success} **MIT** as well
     ([license text](https://github.com/nlohmann/json/blob/develop/LICENSE.MIT)).
 
-@section Trade-TinyGltfImporter-limitations Behavior and limitations
+@section Trade-TinyGltfImporter-usage Usage
 
-The plugin supports @ref Feature::OpenData and @ref Feature::FileCallback
-features. The `tiny_gltf` library loads everything during initial import,
-meaning all external file loading callbacks are called with
-@ref ImporterFileCallbackPolicy::LoadTemporary and the resources can be safely
-freed right after the @ref openData() / @ref openFile() function exits. In case
-of images, the files are loaded on-demand inside @ref image2D() calls with
-@ref ImporterFileCallbackPolicy::LoadTemporary and @ref ImporterFileCallbackPolicy::Close
-is emitted right after the file is fully read.
+This plugin depends on the @ref Trade library and the @ref AnyImageImporter
+plugin and is built if `WITH_TINYGLTFIMPORTER` is enabled when building Magnum
+Plugins. To use as a dynamic plugin, load @cpp "TinyGltfImporter" @ce via
+@ref Corrade::PluginManager::Manager.
+
+Additionally, if you're using Magnum as a CMake subproject, bundle the
+[magnum-plugins repository](https://github.com/mosra/magnum-plugins) and do the
+following:
+
+@code{.cmake}
+set(WITH_ANYIMAGEIMPORTER ON CACHE BOOL "" FORCE)
+add_subdirectory(magnum EXCLUDE_FROM_ALL)
+
+set(WITH_TINYGLTFIMPORTER ON CACHE BOOL "" FORCE)
+add_subdirectory(magnum-plugins EXCLUDE_FROM_ALL)
+
+# So the dynamically loaded plugin gets built implicitly
+add_dependencies(your-app MagnumPlugins::TinyGltfImporter)
+@endcode
+
+To use as a static plugin or as a dependency of another plugin with CMake, put
+[FindMagnumPlugins.cmake](https://github.com/mosra/magnum-plugins/blob/master/modules/FindMagnumPlugins.cmake)
+into your `modules/` directory, request the `TinyGltfImporter` component of the
+`MagnumPlugins` package and link to the `MagnumPlugins::TinyGltfImporter`
+target:
+
+@code{.cmake}
+find_package(MagnumPlugins REQUIRED TinyGltfImporter)
+
+# ...
+target_link_libraries(your-app PRIVATE MagnumPlugins::TinyGltfImporter)
+@endcode
+
+See @ref building-plugins, @ref cmake-plugins, @ref plugins and
+@ref file-formats for more information.
+
+@section Trade-TinyGltfImporter-behavior Behavior and limitations
+
+The plugin supports @ref ImporterFeature::OpenData and
+@ref ImporterFeature::FileCallback features. The `tiny_gltf` library loads
+everything during initial import, meaning all external file loading callbacks
+are called with @ref InputFileCallbackPolicy::LoadTemporary and the resources
+can be safely freed right after the @ref openData() / @ref openFile() function
+exits. In case of images, the files are loaded on-demand inside @ref image2D()
+calls with @ref InputFileCallbackPolicy::LoadTemporary and
+@ref InputFileCallbackPolicy::Close is emitted right after the file is fully
+read.
 
 Import of skeleton, skin and morph data is not supported at the moment.
 
-@subsection Trade-TinyGltfImporter-limitations-animation Animation import
+@subsection Trade-TinyGltfImporter-behavior-animation Animation import
 
 -   Linear quaternion rotation tracks are postprocessed in order to make it
     possible to use the faster
@@ -131,7 +163,7 @@ Import of skeleton, skin and morph data is not supported at the moment.
     however, it can happen that multiple conflicting tracks affecting the same
     node are merged in the same clip, causing the animation to misbehave.
 
-@subsection Trade-TinyGltfImporter-limitations-objects Scene and object import
+@subsection Trade-TinyGltfImporter-behavior-objects Scene and object import
 
 -   If no @cb{.json} "scene" @ce property is present and the file contains at
     least one scene, @ref defaultScene() returns @cpp 0 @ce instead of
@@ -146,29 +178,67 @@ Import of skeleton, skin and morph data is not supported at the moment.
     @cb{.ini} normalizeQuaternions @ce option, see
     @ref Trade-TinyGltfImporter-configuration "below".
 
-@subsection Trade-TinyGltfImporter-limitations-camera Camera import
+@subsection Trade-TinyGltfImporter-behavior-camera Camera import
 
 -   Cameras in glTF are specified with vertical FoV and vertical:horizontal
     aspect ratio, these values are recalculated for horizontal FoV and
     horizontal:vertical aspect ratio as is common in Magnum
 
-@subsection Trade-TinyGltfImporter-limitations-lights Light import
+@subsection Trade-TinyGltfImporter-behavior-lights Light import
 
--   Currently, the importer only supports the historical `KHR_lights_cmn`
-    extension, supported by the original Khronos Blender glTF Exporter.
-    -   Light properties besides light type are not imported.
-    -   Light intensity is not yet supported due to glTF extension draft state.
+-   The importer supports the [KHR_lights_punctual](https://github.com/KhronosGroup/glTF/tree/master/extensions/2.0/Khronos/KHR_lights_punctual)
+    extension, supported by the official glTF Exporter in Blender 2.8.x.
+    -   Only light properties color, intensity and type are imported.
 
-@subsection Trade-TinyGltfImporter-limitations-meshes Mesh import
+@subsection Trade-TinyGltfImporter-behavior-meshes Mesh import
 
--   Meshes with interleaved vertex data/buffer views are not supported.
--   Multi-primitive meshes are loaded as follows:
-    -   The @ref mesh3DCount() query returns a number of all *primitives*, not
+-   Indices are imported as either @ref MeshIndexType::UnsignedByte,
+    @ref MeshIndexType::UnsignedShort or @ref MeshIndexType::UnsignedInt
+-   Positions are imported as @ref VertexFormat::Vector3,
+    @ref VertexFormat::Vector3ub, @ref VertexFormat::Vector3b,
+    @ref VertexFormat::Vector3us, @ref VertexFormat::Vector3s,
+    @ref VertexFormat::Vector3ubNormalized,
+    @ref VertexFormat::Vector3bNormalized,
+    @ref VertexFormat::Vector3usNormalized or
+    @ref VertexFormat::Vector3sNormalized (which includes the additional types
+    specified by [KHR_mesh_quantization](https://github.com/KhronosGroup/glTF/blob/master/extensions/2.0/Khronos/KHR_mesh_quantization/README.md))
+-   Normals, if any, are imported as @ref VertexFormat::Vector3,
+    @ref VertexFormat::Vector3bNormalized or
+    @ref VertexFormat::Vector3sNormalized
+-   Texture coordinates are imported as @ref VertexFormat::Vector3,
+    @ref VertexFormat::Vector3ub, @ref VertexFormat::Vector3b,
+    @ref VertexFormat::Vector3us, @ref VertexFormat::Vector3s,
+    @ref VertexFormat::Vector3ubNormalized,
+    @ref VertexFormat::Vector3bNormalized,
+    @ref VertexFormat::Vector3usNormalized or
+    @ref VertexFormat::Vector3sNormalized (which includes the additional types
+    specified by [KHR_mesh_quantization](https://github.com/KhronosGroup/glTF/blob/master/extensions/2.0/Khronos/KHR_mesh_quantization/README.md)). The
+    data are by default Y-flipped on import unless
+    @cb{.conf} textureCoordinateYFlipInMaterial @ce is either explicitly
+    enabled, or if the file contains non-normalized integer or normalized
+    signed integer texture coordinates (which can't easily be flipped). In that
+    case texture coordinate data are kept as-is and materials provide a texture
+    transformation that does the Y-flip instead.
+-   Colors are imported as @ref VertexFormat::Vector3,
+    @ref VertexFormat::Vector4,
+    @ref VertexFormat::Vector3ubNormalized,
+    @ref VertexFormat::Vector4ubNormalized,
+    @ref VertexFormat::Vector3usNormalized or
+    @ref VertexFormat::Vector4usNormalized
+-   Per-vertex object ID attribute is imported as either
+    @ref VertexFormat::UnsignedInt, @ref VertexFormat::UnsignedShort or
+    @ref VertexFormat::UnsignedByte. By default `_OBJECT_ID` is the recognized
+    name, use the @cb{.ini} objectIdAttribute @ce
+    @ref Trade-TinyGltfImporter-configuration "configuration option" to change
+    the identifier that's being looked for.
+-   Multi-primitive meshes are loaded as follows, consistently with the
+    behavior of @link AssimpImporter @endlink:
+    -   The @ref meshCount() query returns a number of all *primitives*, not
         meshes
-    -   Each multi-primitive mesh is split into a sequence of @ref MeshData3D
+    -   Each multi-primitive mesh is split into a sequence of @ref MeshData
         instances following each other
-    -   @ref mesh3DForName() points to the first mesh in given sequence and
-        @ref mesh3DName() returns the same name for all meshes in given
+    -   @ref meshForName() points to the first mesh in given sequence and
+        @ref meshName() returns the same name for all meshes in given
         sequence
     -   The @ref object3DCount() query returns a number of all nodes extended
         with number of extra nodes for each additional mesh primitive
@@ -184,11 +254,29 @@ Import of skeleton, skin and morph data is not supported at the moment.
         extra nodes, always pointing to the first object in the sequence and
         thus indirectly affecting transformations of the extra nodes
         represented as its children
+-   Attribute-less meshes either with or without an index buffer are supported,
+    however since glTF has no way of specifying vertex count for those,
+    returned @ref Trade::MeshData::vertexCount() is set to @cpp 0 @ce
 
-@subsection Trade-TinyGltfImporter-limitations-materials Material import
+Custom and unrecognized vertex attributes of allowed types are present in the
+imported meshes as well. Their mapping to/from a string can be queried using
+@ref meshAttributeName() and @ref meshAttributeForName(). Attributes with
+unsupported types (such as non-normalized integer matrices) cause the import to
+fail.
 
--   Subset of all material specs is currently imported as @ref PhongMaterialData
+@subsection Trade-TinyGltfImporter-behavior-materials Material import
+
+-   Subset of all material specs is currently imported as @ref PhongMaterialData,
+    including normal textures
 -   Ambient color is always @cpp 0x000000_rgbf @ce (never a texture)
+-   Unless explicitly enabled with the @cb{.ini} allowMaterialTextureCoordinateSets @ce
+    @ref Trade-TinyGltfImporter-configuration "configuration option", only the
+    first set of texture coordinates is supported.
+-   If [KHR_texture_transform](https://github.com/KhronosGroup/glTF/blob/master/extensions/2.0/Khronos/KHR_texture_transform/README.md)
+    is present, the imported material has @ref PhongMaterialData::Flag::TextureTransformation
+    set. All textures have to share the same transformation and the
+    transformation, if present, has to affect the first set of texture
+    coordinates.
 -   For the Metallic/Roughness material spec ([in core](https://github.com/KhronosGroup/glTF/tree/master/specification/2.0#metallic-roughness-material),
     default):
     -   The `baseColorTexture` field is used for diffuse texture, if present.
@@ -208,21 +296,10 @@ Import of skeleton, skin and morph data is not supported at the moment.
         Otherwise, `specularFactor` is used for specular color, if present.
         Otherwise, @cpp 0xffffff_rgbf @ce is used.
     -   Shininess is always @cpp 1.0f @ce
--   For the Blinn/Phong material spec (*draft*
-    [KHR_materials_cmnBlinnPhong](https://github.com/KhronosGroup/glTF/tree/gltf2-common-materials/extensions/Khronos/KHR_materials_cmnBlinnPhong)
-    extension):
-    -   The `diffuseTexture` field is used for diffuse texture, if present.
-        Otherwise, `diffuseFactor` is used for diffuse color, if present.
-        Otherwise, @cpp 0xffffff_rgbf @ce is used.
-    -   The `specularShininessTexture` field is used for specular texture, if
-        present (note that only the RGB channels should be used and the alpha
-        channel --- containing snininess --- should be ignored). Otherwise,
-        `specularFactor` is used for diffuse color, if present. Otherwise,
-        @cpp 0xffffff_rgbf @ce is used.
-    -   The `snininessFactor` field is used for snininess, if present.
-        Otherwise @cpp 1.0f @ce is used.
+-   The [KHR_materials_unlit](https://github.com/KhronosGroup/glTF/blob/master/extensions/2.0/Khronos/KHR_materials_unlit/README.md)
+    extension is currently not supported.
 
-@subsection Trade-TinyGltfImporter-limitations-textures Texture and image import
+@subsection Trade-TinyGltfImporter-behavior-textures Texture and image import
 
 <ul>
 <li>Texture type is always @ref Trade::TextureData::Type::Texture2D, as glTF
@@ -313,7 +390,7 @@ importer-specific data accessors:
     structure
 -   @ref ImageData::importerState() returns pointer to the `tinygltf::Image`
     structure
--   @ref MeshData3D::importerState() returns pointer to the `tinygltf::Mesh`
+-   @ref MeshData::importerState() returns pointer to the `tinygltf::Mesh`
     structure
 -   @ref ObjectData3D::importerState() returns pointer to the `tinygltf::Node`
     structure
@@ -365,7 +442,7 @@ class MAGNUM_TINYGLTFIMPORTER_EXPORT TinyGltfImporter: public AbstractImporter {
     private:
         struct Document;
 
-        MAGNUM_TINYGLTFIMPORTER_LOCAL Features doFeatures() const override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL ImporterFeatures doFeatures() const override;
 
         MAGNUM_TINYGLTFIMPORTER_LOCAL bool doIsOpened() const override;
 
@@ -400,10 +477,14 @@ class MAGNUM_TINYGLTFIMPORTER_EXPORT TinyGltfImporter: public AbstractImporter {
         MAGNUM_TINYGLTFIMPORTER_LOCAL std::string doObject3DName(UnsignedInt id) override;
         MAGNUM_TINYGLTFIMPORTER_LOCAL Containers::Pointer<ObjectData3D> doObject3D(UnsignedInt id) override;
 
-        MAGNUM_TINYGLTFIMPORTER_LOCAL UnsignedInt doMesh3DCount() const override;
-        MAGNUM_TINYGLTFIMPORTER_LOCAL Int doMesh3DForName(const std::string& name) override;
-        MAGNUM_TINYGLTFIMPORTER_LOCAL std::string doMesh3DName(UnsignedInt id) override;
-        MAGNUM_TINYGLTFIMPORTER_LOCAL Containers::Optional<MeshData3D> doMesh3D(UnsignedInt id) override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL UnsignedInt doMeshCount() const override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL Int doMeshForName(const std::string& name) override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL std::string doMeshName(UnsignedInt id) override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL Containers::Optional<MeshData> doMesh(UnsignedInt id, UnsignedInt level) override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL MeshAttribute doMeshAttributeForName(const std::string& name) override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL std::string doMeshAttributeName(UnsignedShort name) override;
+
+        MAGNUM_TINYGLTFIMPORTER_LOCAL bool materialTexture(const char* name, Int texture, Int texCoord, const tinygltf::Value& extensions, UnsignedInt& index, UnsignedInt& coordinateSet, Containers::Optional<Matrix3>& textureMatrix, PhongMaterialData::Flags& flags) const;
 
         MAGNUM_TINYGLTFIMPORTER_LOCAL UnsignedInt doMaterialCount() const override;
         MAGNUM_TINYGLTFIMPORTER_LOCAL Int doMaterialForName(const std::string& name) override;
@@ -415,10 +496,13 @@ class MAGNUM_TINYGLTFIMPORTER_EXPORT TinyGltfImporter: public AbstractImporter {
         MAGNUM_TINYGLTFIMPORTER_LOCAL std::string doTextureName(UnsignedInt id) override;
         MAGNUM_TINYGLTFIMPORTER_LOCAL Containers::Optional<TextureData> doTexture(UnsignedInt id) override;
 
+        MAGNUM_TINYGLTFIMPORTER_LOCAL AbstractImporter* setupOrReuseImporterForImage(UnsignedInt id, const char* errorPrefix);
+
         MAGNUM_TINYGLTFIMPORTER_LOCAL UnsignedInt doImage2DCount() const override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL UnsignedInt doImage2DLevelCount(UnsignedInt id) override;
         MAGNUM_TINYGLTFIMPORTER_LOCAL Int doImage2DForName(const std::string& name) override;
         MAGNUM_TINYGLTFIMPORTER_LOCAL std::string doImage2DName(UnsignedInt id) override;
-        MAGNUM_TINYGLTFIMPORTER_LOCAL Containers::Optional<ImageData2D> doImage2D(UnsignedInt id) override;
+        MAGNUM_TINYGLTFIMPORTER_LOCAL Containers::Optional<ImageData2D> doImage2D(UnsignedInt id, UnsignedInt level) override;
 
         MAGNUM_TINYGLTFIMPORTER_LOCAL const void* doImporterState() const override;
 
